
from datetime import datetime
import re
import sys

from dateutil import parser
import pytz
import six

from . text import levenshtein

UTC_EPOCH = datetime(1970, 1, 1).replace(tzinfo=pytz.utc)
MAX_POSIX_TIMESTAMP = pow(2, 32) - 1

DAYS = {
    'monday',
    'tuesday',
    'wednesday',
    'thursday',
    'friday',
    'saturday',
    'sunday'
}

DAYS_ABBR = [day[:3] for day in DAYS]


class timestamp_ms(object):
    """Build UTC timestamp in milliseconds
    """

    TIMEZONE_PARENTHESIS = re.compile('(.*)\(([a-zA-Z]+)\)$')
    TIMEZONE_SEPARATOR = re.compile('(.*)(\d\d)[.-](\d\d)$')
    QUOTED_TIMEZONE = re.compile("""(.*)['"]([\w:+-]+)['"]?$""")
    START_WITH_DAY_OF_WEEK = re.compile('^([a-zA-Z]*)[\s,](.*)')

    @classmethod
    def feeling_lucky(cls, obj):
        """Tries to convert given object to an UTC timestamp is ms, based
        on its type.
        """
        if isinstance(obj, six.string_types):
            return cls.from_str(obj)
        elif isinstance(obj, six.integer_types) and obj <= MAX_POSIX_TIMESTAMP:
            return cls.from_posix_timestamp(obj)
        elif isinstance(obj, datetime):
            return cls.from_datetime(obj)
        else:
            raise ValueError(
                u"Don't know how to get timestamp from '{}'".format(obj)
            )

    @classmethod
    def fix_mispelled_day(cls, timestr):
        """fix mispelled day when written in english

        :return: `None` if the day was not modified, the new date otherwise
        """
        day_extraction = cls.START_WITH_DAY_OF_WEEK.match(timestr)
        if day_extraction is not None:
            day = day_extraction.group(1).lower()
            if len(day) == 3:
                dataset = DAYS_ABBR
            else:
                dataset = DAYS
            if day not in dataset:
                days = list(dataset)
                days.sort(key=lambda e: levenshtein(day, e))
                return days[0] + day_extraction.group(2)

    @classmethod
    def remove_parenthesis_around_tz(cls, timestr):
        """get rid of parenthesis around timezone: (GMT) => GMT

        :return: the new string if parenthesis were found, `None` otherwise
        """
        parenthesis = cls.TIMEZONE_PARENTHESIS.match(timestr)
        if parenthesis is not None:
            return parenthesis.group(1) + parenthesis.group(2)

    @classmethod
    def remove_quotes_around_tz(cls, timestr):
        """Remove quotes (single and double) around timezone otherwise
        `dateutil.parser.parse` raises
        """
        quoted = cls.QUOTED_TIMEZONE.match(timestr)
        if quoted is not None:
            return quoted.group(1) + quoted.group(2)

    @classmethod
    def remove_timezone(cls, timestr):
        """Completely remove timezone information, if any.

        :return: the new string if timezone was found, `None` otherwise
        """
        if re.match(r".*[\-+]?\d{2}:\d{2}$", timestr):
            return re.sub(
                r"(.*)(\s[\+-]?\d\d:\d\d)$",
                r"\1",
                timestr
            )

    @classmethod
    def fix_timezone_separator(cls, timestr):
        """Replace invalid timezone separator to prevent
        `dateutil.parser.parse` to raise.

        :return: the new string if invalid separators were found,
                 `None` otherwise
        """
        tz_sep = cls.TIMEZONE_SEPARATOR.match(timestr)
        if tz_sep is not None:
            return tz_sep.group(1) + tz_sep.group(2) + ':' + tz_sep.group(3)
        return timestr

    @classmethod
    def from_str(cls, timestr, shaked=False):
        """Use `dateutil` module to parse the give string

        :param basestring timestr: string representing a date to parse
        :param bool shaked: whether the input parameter been already
        cleaned or not.
        """
        orig = timestr
        if not shaked:
            timestr = cls.fix_timezone_separator(timestr)
        try:
            date = parser.parse(timestr)
        except ValueError:
            if not shaked:
                shaked = False
                for shaker in [
                        cls.fix_mispelled_day,
                        cls.remove_parenthesis_around_tz,
                        cls.remove_quotes_around_tz]:
                    new_timestr = shaker(timestr)
                    if new_timestr is not None:
                        timestr = new_timestr
                        shaked = True
                if shaked:
                    try:
                        return cls.from_str(timestr, shaked=True)
                    except ValueError:
                        # raise ValueError below with proper message
                        pass
            msg = u"Unknown string format: {!r}".format(orig)
            raise ValueError(msg), None, sys.exc_info()[2]
        else:
            try:
                return cls.from_datetime(date)
            except ValueError:
                new_str = cls.remove_timezone(orig)
                if new_str is not None:
                    return cls.from_str(new_str)
                else:
                    raise

    @classmethod
    def from_ymd(cls, year, month=1, day=1):
        return cls.from_datetime(datetime(
            year=year, month=month, day=day
        ))

    @classmethod
    def from_posix_timestamp(cls, ts):
        return cls.from_datetime(datetime.utcfromtimestamp(ts))

    @classmethod
    def from_datetime(cls, date):
        if date.tzinfo is None:
            date = date.replace(tzinfo=pytz.utc)
        seconds = (date - UTC_EPOCH).total_seconds() * 1e3
        micro_seconds = date.microsecond / 1e3
        return int(seconds + micro_seconds)

    @classmethod
    def now(cls):
        return cls.from_datetime(datetime.utcnow())



from contextlib import contextmanager

import requests
import json


__all__ = [
    'activate_pyopenssl_for_urllib3',
]


def activate_pyopenssl_for_urllib3():
    """
    Workaround issue described here:
    https://urllib3.readthedocs.org/en/latest/security.html#insecureplatformwarning

    """
    import urllib3.contrib.pyopenssl
    urllib3.contrib.pyopenssl.inject_into_urllib3()


class delayed_request(object):
    """Build streamed file-like instances from an HTTP request.
    """
    def __init__(self, url, method='GET', **kwargs):
        """
        :param basestring url:
          Resource URL

        :param basestring method:
          HTTP method

        :param dict kwargs:
          Optional parameters given to the `requests.sessions.Session.request`
          member method.
        """
        self.__url = url
        self.__method = method
        self.__kwargs = kwargs.copy()

    @contextmanager
    def open(self, session=None):
        """
        :param requests.Session session:
          Optional requests session


        :return:
          file-like object over the decoded bytes
        """
        self.__kwargs.update(stream=True)
        session = session or requests
        resp = session.request(self.__method, self.__url, **self.__kwargs)
        try:
            yield resp
        finally:
            resp.close()

    def __repr__(self):
        return json.dumps(self.__kwargs)



from contextlib import closing
try:
    from cStringIO import StringIO
except ImportError:  # flake8: noqa
    from StringIO import StringIO
import os.path as osp
import mimetypes


class iterator_to_file(object):
    """Incomplete implementation of a file-like object on top of
    an iterator.
    """
    def __init__(self, it):
        self._it = it
        self._buf = ''

    def __iter__(self):
        return self

    def close(self):
        pass

    def read(self, size=-1):
        if size >= 0:
            return self._read_n(size)
        else:
            with closing(StringIO()) as res:
                buf = self._read_n(4096)
                res.write(buf)
                while len(buf) == 4096:
                    buf = self._read_n(4096)
                    res.write(buf)
                return res.getvalue()

    def _read_n(self, size):
        while len(self._buf) < size:
            try:
                data = self.next()
            except StopIteration:
                if len(self._buf) == 0:
                    return ''
                else:
                    res = self._buf
                    self._buf = ''
                    return res
            self._buf += data
            if len(data) == 0:
                break
        if len(self._buf) > size:
            res = self._buf[:size]
            self._buf = self._buf[size:]
        else:
            len(self._buf)
            res = self._buf
            self._buf = ''
        return res

    def next(self):
        return self._it.next()


class FileProperties(object):
    TYPE_CHECK = {
        '3gp': 'video',
        'aaf': 'video',
        'aiff': 'sound',
        'ami': 'document',
        'ape': 'sound',
        'asc': 'document',
        'asf': 'video',
        'ast': 'sound',
        'au': 'sound',
        'avchd': 'video',
        'avi': 'video',
        'bmp': 'image',
        'bwf': 'sound',
        'cdda': 'sound',
        'csv': 'document',
        'doc': 'document',
        'docm': 'document',
        'docx': 'document',
        'dot': 'document',
        'dotx': 'document',
        'epub': 'document',
        'flac': 'sound',
        'flv': 'video',
        'gdoc': 'document',
        'gif': 'image',
        'gslides': 'slide',
        'jpeg': 'image',
        'jpg': 'image',
        'key': 'slide',
        'keynote': 'slide',
        'm4a': 'sound',
        'm4p': 'sound',
        'm4v': 'video',
        'mkv': 'video',
        'mng': 'video',
        'mov': 'video',
        'movie': 'video',
        'mp3': 'sound',
        'mp4': 'video',
        'mpe': 'video',
        'mpeg': 'video',
        'mpg': 'video',
        'nb': 'slide',
        'nbp': 'slide',
        'nsv': 'video',
        'odm': 'document',
        'odp': 'slide',
        'ods': 'document',
        'odt': 'document',
        'ott': 'document',
        'pages': 'document',
        'pdf': 'document',
        'pez': 'slide',
        'png': 'image',
        'pot': 'slide',
        'pps': 'slide',
        'ppt': 'slide',
        'pptx': 'slide',
        'rtf': 'document',
        'sdw': 'document',
        'shf': 'slide',
        'shn': 'sound',
        'show': 'slide',
        'shw': 'slide',
        'swf': 'video',
        'thmx': 'slide',
        'txt': 'document',
        'wav': 'sound',
        'wma': 'sound',
        'wmv': 'video',
        'wpd': 'document',
        'wps': 'document',
        'wpt': 'document',
        'wrd': 'document',
        'wri': 'document',
        'xls': 'document',
        'xlsx': 'document'
    }

    MIMETYPE_CHECK = {
        'image/png': 'image',
        'image/jpeg': 'image',
        'image/jpg': 'image',
        'image/gif': 'image',
        'application/pdf': 'document',
        'application/vnd.google-apps.document': 'document',
        'application/vnd.google-apps.spreadsheet': 'document',
        'application/vnd.google-apps.photo': 'image',
        'application/vnd.google-apps.drawing': 'image',
        'application/vnd.google-apps.presentation': 'slide',
        'application/vnd.google-apps.video': 'video'
    }

    @classmethod
    def file_type(cls, filename, mime_type=None):
        filetype = None
        if filename:
            _, extension = osp.splitext(filename)
            extension = extension[1:]
            filetype = cls.TYPE_CHECK.get(extension)
        if filetype is None:
            filetype = cls.MIMETYPE_CHECK.get(mime_type, 'other')
        return filetype


    @classmethod
    def mime_type(cls, filename):
        mime_type, _ = mimetypes.guess_type(filename)
        return mime_type




import logging
from math import ceil

from docido_sdk.toolbox.collections_ext import chunks

LOGGER = logging.getLogger(__name__)


def check_custom_concurrency(default, forced, logger=None):
    """ Get the proper concurrency value according to the default one
    and the one specified by the crawler.

    :param int default:
      default tasks concurrency

    :param forced:
      concurrency asked by crawler

    :return:
      concurrency to use.
    :rtype: int
    """
    logger = logger or LOGGER
    cmc_msg = 'Invalid "max_concurrent_tasks: '
    if not isinstance(forced, int):
        logger.warn(cmc_msg + 'expecting int')
    elif forced > default:
        msg = 'may not be greater than: %s' % default
        logger.warn(cmc_msg + msg)
    elif forced < 1:
        msg = 'may not be less than 1'
        logger.warn(cmc_msg + msg)
    else:
        default = forced
    return default


def reorg_crawl_tasks(tasks, concurrency, logger=None):
    """ Extract content returned by the crawler `iter_crawl_tasks`
    member method.

    :return:
      tuple made of the sub-tasks to executed, the epilogue task to execute
      or `None` is none was specified by the crawler, and the proper
      tasks concurrency level.
    :rtype: tuple (sub-tasks, epilogue, concurrent)
    """
    futures = tasks['tasks']
    epilogue = tasks.get('epilogue')
    custom_concurrency = tasks.get('max_concurrent_tasks', concurrency)
    check_custom_concurrency(concurrency, custom_concurrency, logger)
    futures = list(futures)
    return futures, epilogue, concurrency


def split_crawl_tasks(tasks, concurrency):
    """ Reorganize tasks according to the tasks max concurrency value.

    :param tasks:
      sub-tasks to execute, can be either a list of tasks of a list of list
      of tasks
    :param int concurrency:
      Maximum number of tasks that might be executed in parallel.

    :return:
      list of list of tasks.
    """
    if any(tasks) and isinstance(tasks[0], list):
        for seq in tasks:
            if not isinstance(seq, list):
                raise Exception("Expected a list of tasks")
    else:
        if concurrency > 1:
            chain_size = int(ceil(float(len(tasks)) / concurrency))
            tasks = [
                chunk for chunk in
                chunks(
                    iter(tasks),
                    max(1, chain_size)
                )
            ]
        else:
            tasks = [tasks]
    return tasks



from contextlib import contextmanager
import datetime
import logging
from argparse import ArgumentParser
import os
import os.path as osp
import pickle
from pickle import PickleError
import shutil
import sys
import time

import six

from .. import loader
from ..env import env
from ..oauth import OAuthToken
from ..core import (
    implements,
    Component,
    ExtensionPoint,
)
from ..crawler import ICrawler
from ..crawler.errors import Retry
from ..index.config import YamlPullCrawlersIndexingConfig
from ..index.processor import (
    Elasticsearch,
    CheckProcessor,
)
from docido_sdk.index.pipeline import IndexPipelineProvider
import docido_sdk.config as docido_config
from ..toolbox.collections_ext import Configuration
from ..toolbox.date_ext import timestamp_ms
from ..crawler.tasks import (
    reorg_crawl_tasks,
    split_crawl_tasks,
)


def wait_or_raise(logger, retry_exc, attempt):
    if attempt == retry_exc.max_retries:
        raise retry_exc
    if retry_exc.countdown is not None:
        assert isinstance(retry_exc.countdown, six.integer_types)
        wait_time = retry_exc.countdown
        if wait_time < 0:
            raise (Exception("'countdown' is less than 0"), None,
                   sys.exc_info()[2])
    else:
        assert isinstance(retry_exc.eta, datetime.datetime)
        target_ts = timestamp_ms.feeling_lucky(retry_exc.eta)
        now_ts = timestamp_ms.now()
        wait_time = (target_ts - now_ts) / 1e3
        if wait_time < 0:
            raise Exception("'eta' is in the future"), None, sys.exc_info()[2]
    logger.warn("Retry raised, waiting {} seconds".format(wait_time))
    time.sleep(wait_time)


def oauth_tokens_from_file():
    path = os.environ.get('DOCIDO_DCC_RUNS', '.dcc-runs.yml')
    crawlers = Configuration.from_env('DOCIDO_CC_RUNS', '.dcc-runs.yml',
                                      Configuration())
    for crawler, runs in crawlers.iteritems():
        for run, run_config in runs.iteritems():
            for k in 'config', 'token':
                if k not in run_config:
                    message = ("In file {}: missing config key '{}'"
                               " in '{}/{}' crawl description.")
                    raise Exception(message.format(path, k, crawler, run))
            if 'config' not in run_config:
                raise Exception("Missing 'config' key")
            run_config.token = OAuthToken(**run_config.token)
    return crawlers


class LocalRunner(Component):
    crawlers = ExtensionPoint(ICrawler)

    def _check_pickle(self, tasks):
        try:
            return pickle.dumps(tasks)
        except PickleError as e:
            raise Exception(
                'unable to serialize crawl tasks: {}'.format(str(e))
            )

    def run(self, logger, config, crawler):
        logger.info("starting crawl")
        self.prepare_crawl_path()
        logger.info('pushed data will be stored in {}'.format(self.crawl_path))
        index_provider = env[IndexPipelineProvider]
        with docido_config:
            if config.config is not None:
                docido_config.clear()
                new_config = Configuration.from_file(config.config)
                docido_config.update(new_config)
            index_api = index_provider.get_index_api(
                self.service, None, None
            )
            attempt = 1
            while True:
                try:
                    tasks = crawler.iter_crawl_tasks(
                        index_api, config.token,
                        logger, config.get('full', False)
                    )
                    break
                except Retry as e:
                    try:
                        wait_or_raise(logger, e, attempt)
                    except:
                        logger.exception('Max retries reached')
                        raise
                    else:
                        attempt += 1
                except Exception:
                    logger.exception('Unexpected exception was raised')
                    raise

            self._check_pickle(tasks)
            tasks, epilogue, concurrency = reorg_crawl_tasks(
                tasks,
                int(config.get('max_concurrent_tasks', 2))
            )
            tasks = split_crawl_tasks(tasks, concurrency)

            def _runtask(task, prev_result):
                attempt = 1
                result = None
                kwargs = dict()
                while True:
                    try:
                        result = task(index_api, config.token,
                                      prev_result, logger, **kwargs)
                        break
                    except Retry as e:
                        try:
                            wait_or_raise(logger, e, attempt)
                        except:
                            logger.exception('Max retries reached')
                            result = e
                            break
                        else:
                            attempt += 1
                            kwargs = e.kwargs
                    except Exception as e:
                        logger.exception('Unexpected exception was raised')
                        result = e
                        break
                return result

            results = []
            for seq in tasks:
                previous_result = None
                for task in seq:
                    previous_result = _runtask(task, previous_result)
                results.append(previous_result)
            if epilogue is not None:
                _runtask(epilogue, results)
        return {
            'service': self.service,
            'name': self.launch,
            'crawl_path': self.crawl_path,
        }

    def get_crawl_path(self):
        now = datetime.datetime.now()
        return osp.join(
            self.crawls_root_path,
            now.strftime('{service}-{launch}-%Y%m%d-%H%M%S'.format(
                service=self.service, launch=self.launch
            ))
        )

    def prepare_crawl_path(self):
        crawl_path = self.get_crawl_path()
        if osp.isdir(crawl_path):
            shutil.rmtree(crawl_path)
        if self.incremental_path is None:
            os.makedirs(crawl_path)
        else:
            parent_crawl_path = osp.dirname(crawl_path)
            if not osp.isdir(parent_crawl_path):
                os.makedirs(parent_crawl_path)
            shutil.copytree(self.incremental_path, crawl_path)
        self.crawl_path = crawl_path

    def run_all(self, crawls):
        crawler_runs = oauth_tokens_from_file()
        for service, launches in crawler_runs.iteritems():
            self.service = service
            c = [c for c in self.crawlers if c.get_service_name() == service]
            if len(c) != 1:
                raise Exception(
                    'unknown crawler for service: {}'.format(service)
                )
            c = c[0]
            for launch, config in launches.iteritems():
                if any(crawls) and launch not in crawls:
                    continue
                self.launch = launch
                logger = logging.getLogger(
                    '{service}.{launch}'.format(service=self.service,
                                                launch=self.launch)
                )
                yield self.run(logger, config, c)


DEFAULT_OUTPUT_PATH = osp.join(os.getcwd(), '.dcc-runs')


def parse_options(args=None):
    if args is None:  # pragma: no cover
        args = sys.argv[1:]
    parser = ArgumentParser()
    parser.add_argument(
        '-i', '--incremental',
        metavar='PATH',
        help='trigger incremental crawl'
    )
    parser.add_argument(
        '-o', '--output',
        metavar='PATH',
        default=DEFAULT_OUTPUT_PATH,
        help='Override persisted data, [default=%(default)s]'
    )
    parser.add_argument(
        '-v', '--verbose',
        action='count',
        dest='verbose',
        help='set verbosity level',
        default=0
    )
    parser.add_argument(
        'crawls',
        metavar='CRAWL',
        nargs='*',
        help='Sub-set of crawls to launch'
    )

    return parser.parse_args(args)


def configure_loggers(verbose):  # pragma: no cover
    logging_level = logging.WARN
    if verbose == 1:
        logging_level = logging.INFO
    elif verbose > 1:
        logging_level = logging.DEBUG
    logging.basicConfig(level=logging_level)
    # shut up a bunch of loggers
    for l in [
        'elasticsearch',
        'requests.packages.urllib3.connectionpool',
        'urllib3.connectionpool',
    ]:
        logging.getLogger(l).setLevel(logging.WARNING)


def _prepare_environment(environment):
    environment = environment or env
    loader.load_components(environment)
    from ..index.test import LocalKV, LocalDumbIndex
    components = [
        YamlPullCrawlersIndexingConfig,
        Elasticsearch,
        CheckProcessor,
        IndexPipelineProvider,
        LocalKV,
        LocalDumbIndex,
    ]
    for component in components:
        _ = environment[component]
        del _  # unused
    return env


@contextmanager
def get_crawls_runner(environment, crawls_root_path, incremental_path):
    from docido_sdk.index.pipeline import IndexAPIConfigurationProvider
    local_runner = None

    class YamlAPIConfigurationProvider(Component):
        implements(IndexAPIConfigurationProvider)

        def get_index_api_conf(self, service, docido_user_id, account_login):
            return {
                'service': service,
                'docido_user_id': docido_user_id,
                'account_login': account_login,
                'local_storage': {
                    'kv': {
                        'path': local_runner.crawl_path,
                    },
                    'documents': {
                        'path': local_runner.crawl_path,
                    }
                }
            }
    environment = _prepare_environment(environment)
    try:
        local_runner = env[LocalRunner]
        local_runner.crawls_root_path = crawls_root_path
        local_runner.incremental_path = incremental_path
        yield local_runner
    finally:
        YamlAPIConfigurationProvider.unregister()


def run(args=None, environment=None):
    args = parse_options(args)
    configure_loggers(args.verbose)
    with get_crawls_runner(environment, args.output,
                           args.incremental) as runner:
        return list(runner.run_all(set(args.crawls)))



import functools
import random
import sys

from requests.exceptions import RequestException

from docido_sdk.toolbox.edsl import kwargsql
from docido_sdk.crawler import Retry

MAX_COLLISIONS = 5
DEFAULT_RETRY = 60


def truncated_exponential_backoff(slot_delay, collision=0, max_collisions=5):
    """Truncated Exponential Backoff
    see https://en.wikipedia.org/wiki/Exponential_backoff
    """
    if collision == max_collisions:
        collision = 0
    slots = random.randint(0, collision)
    return slots * slot_delay


def teb_retry(exc=RequestException,
              when=dict(response__status_code=429),
              delay='response__headers__Retry-After',
              max_collisions=MAX_COLLISIONS,
              default_retry=DEFAULT_RETRY):
    """Decorator catching rate limits exceed events during a crawl task.
    It retries the task later on, following a truncated exponential backoff.
    """
    def wrap(f):
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            attempt = kwargs.pop('teb_retry_attempt', 0)
            try:
                return f(*args, **kwargs)
            except exc as e:
                if kwargsql.and_(e, **when):
                    try:
                        retry_after = kwargsql.get(e, delay)
                    except:
                        retry_after = default_retry
                    else:
                        if retry_after is not None:
                            retry_after = int(retry_after)
                        else:
                            retry_after = default_retry
                    countdown = retry_after + truncated_exponential_backoff(
                        retry_after, attempt % max_collisions)
                    raise Retry(kwargs=dict(attempt=attempt + 1),
                                countdown=countdown)
                else:
                    raise e, None, sys.exc_info()[2] # flake8: noqa.
        return wrapped_f
    return wrap
